with Globals, Strings, IPowers;
use Globals, Strings;

pragma Elaborate_All (Globals,Strings,IPowers);

package body Taylors1.Skews.Pairs is

  package SIP is new IPowers (Scalar => Scalar);

  procedure Show_Param is
    Prec: constant Positive := Get_Precision(SOne/Three);
  begin
    Show1("        PDeg = ",PDeg);
    Show1("        RhoF =",RhoF);
    Show1("        RhoG =",RhoG);
    Show1("Mat_PackSize = ",Mat_PackSize);
    Show1("     QC'Last = ",QC'Last);
    Show0("     RepType =  " & SInfo.RepType);
    Show1("   Precision = ",Prec);
    Show1("      STrunc =  ",STrunc);
    Show1("      NTasks = ",NTasks);
    if Initial_Precision<Prec then raise Sorry; end if; --- just to avoid unpleasant surprises, since
  end Show_Param;                           --- some Scalar values are set during package elaboration

  function IsZero(P: Skew2) return Boolean is
  begin
    return IsZero(P.F) and then IsZero(P.G);
  end IsZero;

  procedure SetZero(R: in Radius2; P: in out Skew2) is
  begin
    SetZero(R.F,P.F);
    SetZero(R.G,P.G);
  end SetZero;

  procedure SetZero(P: in out Skew2) is
  begin
    SetZero(RhoF,P.F);
    SetZero(RhoG,P.G);
  end SetZero;

  procedure Copy(P1: in Skew2; P2: in out Skew2) is
  begin
    Copy(P1.F,P2.F);
    Copy(P1.G,P2.G);
  end Copy;

  procedure Swap(P1,P2: in out Skew2) is
  begin
    Swap(P1.F,P2.F);
    Swap(P1.G,P2.G);
  end Swap;

  procedure Add(P1: in Skew2; P2: in out Skew2) is
  begin
    Add(P1.F,P2.F);
    Add(P1.G,P2.G);
  end Add;

  procedure Sum(P1,P2: in Skew2; P3: in out Skew2) is
  begin
    Sum(P1.F,P2.F,P3.F);
    Sum(P1.G,P2.G,P3.G);
  end Sum;

  procedure Sub(P1: in Skew2; P2: in out Skew2) is
  begin
    Sub(P1.F,P2.F);
    Sub(P1.G,P2.G);
  end Sub;

  procedure Diff(P1,P2: in Skew2; P3: in out Skew2) is
  begin
    Diff(P1.F,P2.F,P3.F);
    Diff(P1.G,P2.G,P3.G);
  end Diff;

  function "-"(P1,P2: Skew2) return Skew2 is
    P3: Skew2;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";

  procedure Mult(S: in Scalar; P: in out Skew2) is
  begin
    Mult(S,P.F);
    Mult(S,P.G);
  end Mult;

  procedure Prod(T: in Taylor1; P1: in Skew2; P2: in out Skew2) is
  begin
    Prod(T,P1.F,P2.F);
    Prod(T,P1.G,P2.G);
  end Prod;

  procedure Mult(T: in Taylor1; P: in out Skew2) is
  begin
    Mult(T,P.F);
    Mult(T,P.G);
  end Mult;

  procedure SameRho(P: in out Skew2) is
  begin
    SameRho(P.F);
    SameRho(P.G);
  end SameRho;

  procedure SetRho(R: in Radius2; P: in out Skew2; Check: in Boolean := True) is
  begin
    SetRho(R.F,P.F,Check);
    SetRho(R.G,P.G,Check);
  end SetRho;

  function MaxNorm(R: Radius2; P: Skew2) return Radius is
  begin
    return MaxNorm(R.F,P.F)+MaxNorm(R.G,P.G);
  end MaxNorm;

  function MaxNorm(P: Skew2) return Radius is
  begin
    return MaxNorm(RhoF,P.F)+MaxNorm(RhoG,P.G);
  end MaxNorm;

  function MaxDist(R: Radius2; P1,P2: Skew2) return Radius is
  begin
    return MaxDist(R.F,P1.F,P2.F)+MaxDist(R.G,P1.G,P2.G);
  end MaxDist;

  function MaxDist(P1,P2: Skew2) return Radius is
  begin
    return MaxDist(RhoF,P1.F,P2.F)+MaxDist(RhoG,P1.G,P2.G);
  end MaxDist;

  ------------------------------------
  ------------------------------------
  ------------------------------------

  procedure Scale1(P: in out Skew2) is
  begin
    Scale1(P.F);
    Scale1(P.G);
  end Scale1;

  procedure Scale3(P: in out Skew2) is
  begin
    Scale3(P.F);
    Scale3(P.G);
  end Scale3;

  procedure Fix_Reversible(P: in out Skew2) is
  begin
    Fix_Reversible(P.F);
    Fix_Reversible(P.G);
  end Fix_Reversible;

  procedure Show_Reversible(N: in String; P: in out Skew2) is
  begin
    Show_Reversible(N & "F ",P.F);
    Show_Reversible(N & "G ",P.G);
  end Show_Reversible;

  procedure Show_DetMinus1(N: in String; P: in Skew2; R: in Radius2 := RhoP) is
  begin
    Show1(N & "||DetF-1|| ~",DetMinus1(P.F,R.F));
    Show1(N & "||DetG-1|| ~",DetMinus1(P.G,R.G));
  end Show_DetMinus1;

  procedure WarnDet1(N: in String; P: in Skew2) is
  begin
    if not P.F.Det1 then Show0(N & "Warning: F.Det1 is False"); end if;
    if not P.G.Det1 then Show0(N & "Warning: G.Det1 is False"); end if;
  end WarnDet1;

  procedure AddBall(S: in Scalar; P: in out Skew2) is
  begin
    AddBall(RhoF,S,P.F);
    AddBall(RhoG,S,P.G);
  end AddBall;

  procedure NumTrunc(P: in out Skew2; Deg: in Power := PDeg; Cut: in Radius := Zero) is
  begin
    NumTrunc(P.F,Deg,Cut);
    NumTrunc(P.G,Deg,Cut);
  end NumTrunc;

  procedure NumTrunc(P1: in Skew2; P2: in out Skew2; Deg: in Power := PDeg; Cut: in Radius := Zero) is
  begin
    Copy(P1,P2);
    NumTrunc(P2,Deg,Cut);
  end NumTrunc;

  procedure ShowGuessRho(N: in String; P: in Skew2) is
  begin
    Show1(N & "guessed RhoF =",GuessRho(P.F));
    Show1(N & "guessed RhoG =",GuessRho(P.G));
  end ShowGuessRho;

  procedure ShowType1(N: in String; P: in Skew2) is
  begin
    ShowType1(N & "F ",P.F);
    ShowType1(N & "G ",P.G);
  end ShowType1;

  procedure ShowType2(N: in String; P1,P2: in Skew2) is
  begin
    ShowType2(N & "F ",P1.F,P2.F);
    ShowType2(N & "G ",P1.G,P2.G);
  end ShowType2;

  procedure ShowNorms(N: in String; P: in Skew2) is
  begin
    ShowNorms(N & "F ",P.F);
    ShowNorms(N & "G ",P.G);
  end ShowNorms;

  procedure Show1(N: in String; P: in Skew2) is
  begin
    Show1(N & "F ",P.F);
    Show1(N & "G ",P.G);
  end Show1;

  procedure Show2(N: in String; P1,P2: in Skew2) is
  begin
    Show2(N & "F ",P1.F,P2.F);
    Show2(N & "G ",P1.G,P2.G);
  end Show2;

  procedure Put(F: in File_Type; P: in Skew2; Decimal: in Boolean := False) is
  begin
    Put(F,P.F,Decimal);
    Put(F,P.G,Decimal);
  end Put;

  procedure Get(F: in File_Type; P: in out Skew2; Decimal: in Boolean := False) is
  begin
    Get(F,P.F,Decimal);
    Get(F,P.G,Decimal);
  end Get;

  procedure Write(FileName: in String; P: in Skew2; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out Skew2; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

  ------------------------------------
  ------------------------------------
  ------------------------------------

  procedure RevCosAM(E: in Scalar; P: in out Skew2) is
  begin
    FZero(P.F,RhoF);
    RevCosAM(E,P.G,RhoG);
  end RevCosAM;

  procedure Normalize(P1: in Skew2; P2: in out Skew2) is
  begin
    Normalize(P1.F,P2.F);
    Normalize(P1.G,P2.G);
  end Normalize;

  procedure Normalize(P: in out Skew2; Force: in Boolean := False) is
    Tmp: Skew;
  begin
    if Force then
      P.F.Det1 := False;
      P.G.Det1 := False;
    end if;
    Normalize(P.F,Tmp);
    Swap(Tmp,P.F);
    Normalize(P.G,Tmp);
    Swap(Tmp,P.G);
  end Normalize;

  procedure RG1(P1: in Skew2; P2: in out Skew2) is
    R: constant Radius := Half+IG1*RhoG;
    UFac,VFac: Scalar;
    Tmp: Skew2;
    Gi: Skew renames Tmp.F;
    FGi: Skew renames Tmp.G;
  begin
    if Verbosity>1 then WarnDet1("",P1); end if;
    Copy(P1.G,P2.F);
    Inv(P1.G,Gi);
    Prod(P1.F,Gi,FGi,R);
    Copy(FGi,P2.G);
    Scale1(P2);
    Equalize(P2.F,P2.G,UFac,VFac);
    if Verbosity>0 then Show_DetMinus1("",P2); end if;
    Normalize(P2,Tmp);
    Swap(Tmp,P2);
  end RG1;

  procedure ProdGFiG(F,G: in Skew; O: in out PData; H: in out Skew; R: in Radius) is
    GFiG: Skew renames H;
    Fi: Skew;
  begin
    Inv(F,Fi);
    Prod_GFG(Fi,G,O,GFiG,R);
  end ProdGFiG;

  procedure ProdGiFGiFGi(F,G: in Skew; O1,O2: in out PData; H: in out Skew; R1,R2: in Radius) is
    GiFGiFGi: Skew renames H;
    Gi,FGiF: Skew;
  begin
    Inv(G,Gi);
    Prod_GFG(Gi,F,O1,FGiF,R1);
    Prod_GFG(FGiF,Gi,O2,GiFGiFGi,R2);
  end ProdGiFGiFGi;

  procedure Normalize(P1: in Skew2; P2: in out Skew2; DFac: in out Skew) is
    M: TMat2 renames DFac.L.all.Data;
    Fac1: Taylor1 renames M(1);
    Fac3: Taylor1 renames M(2);
    Gac1: Taylor1 renames M(3);
    Gac3: Taylor1 renames M(4);
  begin
    DetFactors(P1.F,Fac1,Fac3);
    DetFactors(P1.G,Gac1,Gac3);
    Normalize(P1.F,Fac1,P2.F);
    Normalize(P1.G,Gac1,P2.G);
  end Normalize;

  procedure RG3(P1: in Skew2; O: in out RGData; P2: in out Skew2; FillO: in Boolean := True) is
    --- FillO also computes/saves O.DFac
    OO: PData renames O.O;
    O1: PData renames O.O1;
    O2: PData renames O.O2;
    HE: Skew2 renames O.HE;
    HN: Skew2 renames O.HN;
  begin
    if Verbosity>1 then WarnDet1("",P1); end if;
    ProdGFiG(P1.F,P1.G,OO,HE.F,RhoHatF);
    ProdGiFGiFGi(P1.F,P1.G,O1,O2,HE.G,RhoFGiF,RhoHatG);
    Scale3(HE);
    if Verbosity>2 then Show_Reversible("",HE); end if;
    Fix_Reversible(HE);
    Copy(HE,HN);
    Equalize(HN.F,HN.G,O.UFac,O.VFac);
    if FillO then
      Normalize(HN,P2,O.DFac);
    else
      Normalize(HN,P2);
    end if;
    Fix_Reversible(P2);
  end RG3;

  procedure RG3(P1: in Skew2; P2: in out Skew2; UFac,VFac: in out Scalar) is
    O: RGData;
  begin
    RG3(P1,O,P2,False);
    Copy(O.UFac,UFac);
    Copy(O.VFac,VFac);
  end RG3;

  procedure RG3(P1: in Skew2; P2: in out Skew2) is
    O: RGData;
  begin
    RG3(P1,O,P2,False);
  end RG3;

  procedure DProdGFiG(DP: in Skew2; O: in PData; DH: in out Skew) is
    DF: Skew renames DP.F;
    DG: Skew renames DP.G;
    D_GFiG: Skew renames DH;
    DFi: Skew;
  begin
    Inv(DF,DFi);
    DProd_GFG(DFi,DG,O,D_GFiG);
  end DProdGFiG;

  procedure DProdGiFGiFGi(DP: in Skew2; O1,O2: in PData; DH: in out Skew) is
    DF: Skew renames DP.F;
    DG: Skew renames DP.G;
    D_GiFGiFGi: Skew renames DH;
    DGi,D_FGiF: Skew;
  begin
    Inv(DG,DGi);
    DProd_GFG(DGi,DF,O1,D_FGiF);
    DProd_GFG(D_FGiF,DGi,O2,D_GiFGiFGi);
  end DProdGiFGiFGi;

  procedure DFactors(O: in RGData; DF: in Skew; DUFac,DVFac: in out Scalar) is
    --- factors used in DEqualize
    M: TMat2 renames O.HE.F.L.all.Data; -- O.HE.F pre-equalized F
    DM: TMat2 renames DF.L.all.Data;
    U0,V0,DU0,DV0,S: Scalar;
  begin
    Val(M(JU),SZero,U0);
    Val(M(JV),SZero,V0);
    Val(DM(JU),SZero,DU0);
    Val(DM(JV),SZero,DV0);
    Quot(DU0,U0,DUFac);
    Quot(DV0,V0,DVFac);
    Diff(DVFac,DUFac,S);
    Mult(SHalf,S);
    Prod(S,O.UFac,DUFac); -- DUFac := UFac*[DV0/V0-DU0/U0]/2
    Neg(S);
    Prod(S,O.VFac,DVFac); -- DVFac := VFac*[DU0/U0-DV0/V0]/2
  end DFactors;

  procedure DEqualize(UFac,VFac,DUFac,DVFac: in Scalar; H1,DH1: in Skew; DH2: in out Skew) is
    M1: TMat2 renames H1.L.all.Data;
    DM1: TMat2 renames DH1.L.all.Data;
    DM2: TMat2 renames DH2.L.all.Data;
  begin
    if DH1.Alpha /= H1.Alpha then raise Sorry; end if;
    DH2.Det1 := False;
    DH2.Alpha := DH1.Alpha;
    Copy(DM1(JT),DM2(JT));
    Copy(DM1(JS),DM2(JS));
    Prod(UFac,DM1(JU),DM2(JU));
    AddProd(DUFac,M1(JU),DM2(JU)); -- D(UFac*U)=DUFac*U+UFac*DU
    Prod(VFac,DM1(JV),DM2(JV));
    AddProd(DVFac,M1(JV),DM2(JV)); -- D(VFac*V)=DVFac*V+VFac*DV
  end DEqualize;

  procedure DEqualize(O: in RGData; DP1: in Skew2; DP2: in out Skew2) is
    DUFac,DVFac: Scalar;
  begin
    DFactors(O,DP1.F,DUFac,DVFac);
    DEqualize(O.UFac,O.VFac,DUFac,DVFac,O.HE.F,DP1.F,DP2.F); -- O.HE.F pre-equalized F
    DEqualize(O.UFac,O.VFac,DUFac,DVFac,O.HE.G,DP1.G,DP2.G); -- O.HE.G pre-equalized G
  end DEqualize;

  procedure DNormalize(O: in RGData; DP1: in Skew2; DP2: in out Skew2) is
    M: TMat2 renames O.DFac.L.all.Data;
    Fac1: Taylor1 renames M(1);
    Fac3: Taylor1 renames M(2);
    Gac1: Taylor1 renames M(3);
    Gac3: Taylor1 renames M(4);
  begin
    DNormalize(O.HN.F,DP1.F,Fac1,Fac3,DP2.F); -- O.HN.F pre-normalized F
    DNormalize(O.HN.G,DP1.G,Gac1,Gac3,DP2.G); -- O.HN.G pre-normalized G
  end DNormalize;

  procedure DRG3(O: in RGData; DP1: in Skew2; DP2: in out Skew2) is
    OO: PData renames O.O;
    O1: PData renames O.O1;
    O2: PData renames O.O2;
    DQ: Skew2;
  begin
    DProdGFiG(DP1,OO,DP2.F);
    DProdGiFGiFGi(DP1,O1,O2,DP2.G);
    SameRho(DP2);
    Scale3(DP2);
    Fix_Reversible(DP2);
    DEqualize(O,DP2,DQ);
    DNormalize(O,DQ,DP2);
    Fix_Reversible(DP2);
  end DRG3;

  -----------------------------------
  -----------------------------------
  -----------------------------------

  procedure Norm(M: in CMode; S: in out Scalar) is
    use SIP;
  begin
    if M.Err then
      Copy(SOne,S);  --- error modes have norm 1
    elsif M.F then
      IPower(M.I,Scal(RhoF),S);
    else
      IPower(M.I,Scal(RhoG),S);
    end if;
  end Norm;

  procedure MaxNorm(M: in CMode; S: out Radius) is
  begin
    if M.Err then
      S := One;
    elsif M.F then
      S := RhoFPow(M.I);
    else
      S := RhoGPow(M.I);
    end if;
  end MaxNorm;

  procedure Assign(M: in CMode; P: in out Skew2) is
  begin
    SetZero(P);           --- with Det1=False
    P.F.Alpha := GOne;
    P.G.Alpha := Inv_Gold;
    AddProd(M,SOne,P,True);
  end Assign;

  procedure AddProd(M: in CMode; S: in Scalar; P: in out Skew2; Check: in Boolean) is
    pragma Unreferenced(Check);
    FL: TMat2 renames P.F.L.all.Data;
    GL: TMat2 renames P.G.L.all.Data;
    R: Radius;
  begin
    if M.F then
      if P.F.Alpha /= GOne then
        if not IsZero(P.F) then raise Undefined; end if;
        P.F.Alpha := GOne;
      end if;
      if M.Err then
        R := InvRhoFPow(M.I)*MaxNorm(S);
        Add_Ball(M.I,R,FL(M.J));
      else
        Add_Coeff(M.I,S,FL(M.J));
      end if;
    else
      if P.G.Alpha /= Inv_Gold then
        if not IsZero(P.G) then raise Undefined; end if;
        P.G.Alpha := Inv_Gold;
      end if;
      if M.Err then
        R := InvRhoGPow(M.I)*MaxNorm(S);
        Add_Ball(M.I,R,GL(M.J));
      else
        Add_Coeff(M.I,S,GL(M.J));
      end if;
    end if;
  end AddProd;

  procedure Extract_Coeff(M: in CMode; P: in out Skew2; S: in out Scalar) is
    --- assuming M.Err=False
    --- assuming M and C are of the same type
    --- S encloses the coefficient
    FL: TMat2 renames P.F.L.all.Data;
    GL: TMat2 renames P.G.L.all.Data;
  begin
    if M.F then
      Extract_Coeff(M.I,FL(M.J),S);
    else
      Extract_Coeff(M.I,GL(M.J),S);
    end if;
  end Extract_Coeff;

  procedure Extract_Err(M: in CMode; P: in out Skew2; S: in out Scalar) is
    --- assuming M.Err=True
    --- assuming M and C are of the same type
    --- S enloses an upper bound on the norm
    FL: TMat2 renames P.F.L.all.Data;
    GL: TMat2 renames P.G.L.all.Data;
  begin
    if M.F then
      Norm1(RhoF,FL(M.J),S);  --- err modes have norm 1
      SetZero(FL(M.J));
    else
      Norm1(RhoG,GL(M.J),S);
      SetZero(GL(M.J));
    end if;
  end Extract_Err;

  procedure Extract(M: in CMode; P: in out Skew2; S: in out Scalar; Check: in Boolean) is
    pragma Unreferenced(Check);
  begin
    if M.Err then
      Extract_Err(M,P,S);
      --- last step, if Not_STrunc
      if M.J=4 and then (not M.F) and then (not IsZero(P)) then raise Sorry; end if;
    else
      Extract_Coeff(M,P,S);
    end if;
  end Extract;

  procedure Show(M: in CMode) is
    MF: constant Character := Choose(M.F,'F','G');
    ME: constant Character := Choose(M.Err,'e','c');
  begin
    Show1(MF & ME & " " & CName(M.J),M.I);
  end Show;

  function Reversible(M: CMode) return Boolean is
  begin
    if M.J=JS then return (M.I mod 2)=1; end if;
    return (M.I mod 2)=0;
  end Reversible;

  procedure Show(M: in CModes) is
  begin
    if M'First /= 1 then raise Inconsistent_Data; end if;
    for K in M'Range loop
      Show0(Strng(K,4) & " ",False);
      Show(M(K));
    end loop;
  end Show;

  function Make(MaxDeg: Power; Trunc: Boolean) return CModes is
    Include_Errs: constant Boolean := not (Trunc or STrunc);

    procedure Build(M: out CModes; Count: out Natural) is
      procedure Set_Mode(I,J: in Natural; F,Err: in Boolean) is
        MC: constant CMode := (I,J,F,Err);
      begin
        if Reversible(MC) then
          Count := Count+1;
          if Count <= M'Last then M(Count) := MC; end if;
        end if;
      end Set_Mode;
    begin
      Count := 0;
      for I in 0 .. MaxDeg loop
        for J in Index4 loop
          Set_Mode(I,J,True,False);  -- F coeff
          Set_Mode(I,J,False,False); -- G coeff
        end loop;
      end loop;
      if Include_Errs then
        for J in Index4 loop
          Set_Mode(MaxDeg,J,True,True);  -- F err
          Set_Mode(MaxDeg,J,False,True); -- G err
        end loop;
      end if;
    end Build;
    Count: Natural;

  begin
    declare
      M: CModes(1..1);
    begin
      Build(M,Count);
    end;
    declare
      M: CModes(1..Count);
    begin
      Build(M,Count);
      return M;
    end;
  end Make;

  ------------------------------------
  ------------------------------------
  ------------------------------------

  procedure Extra_Info(P: in Skew2; UFac,VFac: in Scalar) is
  begin
    Show1("UFac",UFac);
    Show1("VFac",VFac);
    ShowCommutator(P.F,P.G);
    ShowEigenA10(P.G,UFac,VFac);
  end Extra_Info;

  procedure DContrNorm(M: in SMatrix; O: in RGData; PackSize: in Positive; R: out Radius) is
    Q: constant CModes := Make(PackSize,STrunc); --- Err modes (but Extract is not used)
    procedure DG(H: in Skew2; DGH: in out Skew2) is begin DRG3(O,H,DGH); end DG;
    procedure DContract is new DContr (DMap => DG);
    procedure DC(H: in Skew2; DCH: in out Skew2) is begin DContract(QC,M,H,DCH); end DC;
    procedure DCNorm is new Op_Norm (LinOp => DC);
  begin
    if PackSize <= Mat_PackSize then raise Not_Implemented; end if;
    if Verbosity>1 then
      Show1("DContrNorm: Q'Last = ",Q'Last);
    end if;
    DCNorm(Q,R);
  end DContrNorm;

  procedure DContrNorm(M: in SMatrix; P: in Skew2; PackSize: in Positive; R: out Radius) is
    Tmp: Skew2;
    O: RGData;
  begin
    RG3(P,O,Tmp);
    Extra_Info(P,O.UFac,O.VFac);
    DContrNorm(M,O,PackSize,R);
  end DContrNorm;

  procedure ContrFix(M: in SMatrix; PackSize: in Positive; P: in out Skew2) is
    --KMax: constant Radius := Flt(7)/Flt(8);
    KMax: constant Radius := Flt(3)/Flt(4);
    E,K: Radius;
    ANorm,R,Eps,Del: Scalar;
    P0: Skew2;
  begin
    Show0("ContrFix: ",False);
    NumTrunc(P,P0);
    RG3(P0,P);
    Sub(P0,P);
    E := MaxNorm(RhoP,P); --- RhoP  new
    Assign(E,Eps);
    OpNorm(QC,M,ANorm);
    Add(SOne,ANorm);
    Show2("Eps,ANorm ",E,Sup(ANorm));
    ---
    Del := Eps/(SOne-Scal(KMax))+Scal(Tiny);
    R := ANorm*Del;
    Copy(P0,P);
    AddBall(R,P);
    DContrNorm(M,P,PackSize,K);
    Show1("ContrFix: K =",K);
    if K>KMax then raise Sorry with "K too large"; end if;
    ---
    Del := Eps/(SOne-Scal(K))+Scal(Tiny);
    R := ANorm*Del;
    Show2("ContrFix: Del,R: ",Sup(Del),Sup(R));
    Copy(P0,P);
    AddBall(R,P);
  end ContrFix;

  function Fix_Name return String is
  begin
    if STrunc then return "approx-Fix." & Strng(PDeg); end if;
    return "Fix." & Strng(PDeg);
  end Fix_Name;

  function Eigen1_Name return String is
  begin
    return "approx-Eigen1." & Strng(PDeg);
  end Eigen1_Name;

  function ContrMat_Name return String is
  begin
    return "ContrMat." & Strng(QC'Last);
  end ContrMat_Name;

  function DerMatrix_Name return String is
  begin
    return "DerMatrix." & Strng(QC'Last);
  end DerMatrix_Name;

  procedure NumTrunc(M: in out SMatrix; Cut: in Radius := Zero) is
    PosCut: constant Boolean := Cut>Zero;
  begin
    for I in M'Range(1) loop
      for J in M'Range(2) loop
        Center(M(I,J));
        if PosCut and then MaxNorm(M(I,J))<Cut then SetZero(M(I,J)); end if;
      end loop;
    end loop;
  end NumTrunc;

  procedure DerMatrix(O: in RGData; M: in out SMatrix) is
    --- for numeric mode only
    procedure DG(H: in Skew2; DGH: in out Skew2) is begin DRG3(O,H,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if Not_STrunc then raise Sorry; end if;
    if Verbosity>1 then Show0("running DerMatrix"); end if;
    Enclose(QC,M);
  end DerMatrix;

  procedure DerMatrix(P: in Skew2; M: in out SMatrix) is
    Tmp: Skew2;
    O: RGData;
  begin
    RG3(P,O,Tmp);
    DerMatrix(O,M);
  end DerMatrix;

  procedure ContrMatrix(O: in RGData; M: in out SMatrix) is
  begin
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    DerMatrix(O,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure ContrMatrix(P: in Skew2; M: in out SMatrix) is
  begin
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    DerMatrix(P,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure NumInterpolate(P0: in Skew2; P1: in out Skew2; Fac: in Flt) is
    --- P1 := (1-Fac)*P0+Fac*P1
  begin
    NumInterpolate(P0.F,P1.F,Fac);
    NumInterpolate(P0.G,P1.G,Fac);
  end NumInterpolate;

  procedure NumIProd(P1,P2: in Skew2; S: in out Scalar) is
    N: Scalar;
  begin
    NumIProd(P1.F,P2.F,S);
    NumIProd(P1.G,P2.G,N);
    Add(N,S);
  end NumIProd;

  procedure AfterContr0(P: in out Skew2) is
    Tmp: Skew2;
  begin
    Fix_Reversible(P);
    Normalize(P,Tmp);
    Swap(Tmp,P);
  end AfterContr0;

  procedure ApproxFix(P: in out Skew2; M: in out SMatrix; Eps: in Radius := Zero) is
    --- find approximate fixed point
    procedure Contract0 is new Contr0 (Map => RG3);
    Iter: constant Integer := 1024;
    MaxCredit: constant Integer := 8;  --- was 4
    Fac: constant Radius := One;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := 1.0E+9;
    Q0,Q: Skew2;
    use SM;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running ApproxFix"); end if;
    Copy(P,Q);
    for I in 1 .. Iter loop
      Copy(Q,Q0);
      ContrMatrix(Q,M);
      Contract0(QC,M,Q,Err);
      Show1("ApproxFix: step " & Strng(I) & ", Err ",Err);
      NumInterpolate(Q0,Q,Fac);  --- Q := (1-Fac)*Q0+Fac*Q
      AfterContr0(Q);            --- new
      if Err<MinErr then
        Write(Output_Dir & Fix_Name,Q0); -- new
        Write(Output_Dir & ContrMat_Name,M); -- new
        Copy(Q0,P);
        MinErr := Err;
        if Err<Eps then exit; end if;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    Show1("ApproxFix: MinErr ",MinErr);
    ContrMatrix(P,M);
  end ApproxFix;

  procedure IterContr(P: in out Skew2; M: in out SMatrix) is
    --- improve approximate fixed point
    procedure Contract0 is new Contr0 (Map => RG3);
    MaxCredit: constant Integer := 16;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    Q: Skew2;
    use SM;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running IterContr"); end if;
    Copy(P,Q);
    ContrMatrix(Q,M);
    while Credit>0 loop
      Contract0(QC,M,Q,Err);
      AfterContr0(Q);               --- new
      Show1("IterContr: Err ",Err);
      if Err<MinErr then
        Copy(Q,P);
        MinErr := Err;
        Credit := MaxCredit;
        ContrMatrix(Q,M);
        Write(Output_Dir & Fix_Name,Q); -- new
        Write(Output_Dir & ContrMat_Name,M); -- new
      else
        Credit := Credit-1;
      end if;
    end loop;
    Show1("IterContr: Err ",MinErr);
    ContrMatrix(P,M);
  end IterContr;

  procedure ApproxNormalize(P: in out Skew2; Fac: in out Scalar) is
    S: Scalar;
  begin
    NumIProd(P,P,S);
    Sqrt(S,Fac);
    Inv(Fac,S);
    Mult(S,P);
  end ApproxNormalize;

  procedure ApproxEigen1(O: in RGData; H: in out Skew2; Eps: in Radius := Zero) is
    --- for numeric mode only
    Iter: constant Integer := 1024;
    Err: Radius;
    EV: Scalar;
    H0: Skew2;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running ApproxEigen1"); end if;
    ApproxNormalize(H,EV);
    for I in 1 .. Iter loop
      Copy(H,H0);
      DRG3(O,H0,H);
      ApproxNormalize(H,EV);
      Show1("ApproxEigen1: step " & Strng(I) & ",  EV ",EV);
      Err := MaxDist(H0,H);
      Show1("ApproxEigen1: step " & Strng(I) & ", Err ",Err);
      Write(Output_Dir & Eigen1_Name,H);
      exit when Err<Eps;
    end loop;
  end ApproxEigen1;

  procedure ApproxEigen1(P: in Skew2; H: in out Skew2; Eps: in Radius := Zero) is
    Tmp: Skew2;
    O: RGData;
  begin
    RG3(P,O,Tmp);
    ApproxEigen1(O,H,Eps);
  end ApproxEigen1;

  procedure QC_UnPack(V: in QC_Vector; P: in out Skew2) is
  begin
    SetZero(P);
    P.F.Alpha := GOne;
    P.G.Alpha := Inv_Gold;
    for K in QC'Range loop
      AddProd(QC(K),V(K),P,False);
    end loop;
  end QC_UnPack;

end Taylors1.Skews.Pairs;
