with ScalVectors, Linear, Linear.Contr;

pragma Elaborate_All (ScalVectors,Linear.Contr);

generic

  Mat_PackSize: in Positive;
  RhoF,RhoG: in Radius;

  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);

package Taylors1.Skews.Pairs is --- for pairs of skew-product maps

  type Radius2 is
    record
      F,G: Radius;
    end record;

  RhoP: constant Radius2 := (RhoF,RhoG);

  type Skew2 is
    record
      F,G: Skew;
    end record;

  procedure Show_Param;
  function IsZero(P: Skew2) return Boolean;
  procedure SetZero(R: in Radius2; P: in out Skew2); --- Alpha=0
  procedure SetZero(P: in out Skew2);                --- Alpha=0, RhoF, RhoG
  procedure Copy(P1: in Skew2; P2: in out Skew2);
  procedure Swap(P1,P2: in out Skew2);
  procedure Add(P1: in Skew2; P2: in out Skew2);
  procedure Sum(P1,P2: in Skew2; P3: in out Skew2);
  procedure Sub(P1: in Skew2; P2: in out Skew2);
  procedure Diff(P1,P2: in Skew2; P3: in out Skew2);
  function "-"(P1,P2: Skew2) return Skew2;
  procedure Mult(S: in Scalar; P: in out Skew2);
  procedure Prod(T: in Taylor1; P1: in Skew2; P2: in out Skew2);
  procedure Mult(T: in Taylor1; P: in out Skew2);
  procedure SameRho(P: in out Skew2);
  procedure SetRho(R: in Radius2; P: in out Skew2; Check: in Boolean := True);
  function MaxNorm(R: Radius2; P: Skew2) return Radius;
  function MaxNorm(P: Skew2) return Radius;
  function MaxDist(R: Radius2; P1,P2: Skew2) return Radius;
  function MaxDist(P1,P2: Skew2) return Radius;

  procedure Scale1(P: in out Skew2);
  procedure Scale3(P: in out Skew2);
  procedure Fix_Reversible(P: in out Skew2);
  procedure Show_Reversible(N: in String; P: in out Skew2);
  procedure Show_DetMinus1(N: in String; P: in Skew2; R: in Radius2 := RhoP);
  procedure WarnDet1(N: in String; P: in Skew2);
  procedure AddBall(S: in Scalar; P: in out Skew2);
  procedure NumTrunc(P: in out Skew2; Deg: in Power := PDeg; Cut: in Radius := Zero);
  procedure NumTrunc(P1: in Skew2; P2: in out Skew2; Deg: in Power := PDeg; Cut: in Radius := Zero);

  procedure ShowGuessRho(N: in String; P: in Skew2);
  procedure ShowType1(N: in String; P: in Skew2);
  procedure ShowType2(N: in String; P1,P2: in Skew2);
  procedure ShowNorms(N: in String; P: in Skew2);
  procedure Show1(N: in String; P: in Skew2);
  procedure Show2(N: in String; P1,P2: in Skew2);
  procedure Put(F: in File_Type; P: in Skew2; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out Skew2; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in Skew2; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out Skew2; Decimal: in Boolean := False);

  -----------------------------------------
  -----------------------------------------
  -----------------------------------------

  procedure RevCosAM(E: in Scalar; P: in out Skew2);
  procedure Normalize(P1: in Skew2; P2: in out Skew2);
  procedure Normalize(P: in out Skew2; Force: in Boolean := False);
  procedure RG1(P1: in Skew2; P2: in out Skew2);

  procedure ProdGFiG(F,G: in Skew; O: in out PData; H: in out Skew; R: in Radius);
  procedure ProdGiFGiFGi(F,G: in Skew; O1,O2: in out PData; H: in out Skew; R1,R2: in Radius);
  procedure Normalize(P1: in Skew2; P2: in out Skew2; DFac: in out Skew);

  type RGData is --- written by RG3, for use by DRG3
    record
      O: PData;           -- for DProdGFiG
      O1,O2: PData;       -- for DProdGiFGiFGi
      HE: Skew2;          -- for DEqualize, pre-equalized
      UFac,VFac: Scalar;  -- for DEqualize
      HN: Skew2;          -- for DNormalize, pre-normalized
      DFac: Skew;         -- for DNormalize
    end record;

  procedure RG3(P1: in Skew2; O: in out RGData; P2: in out Skew2; FillO: in Boolean := True);
  procedure RG3(P1: in Skew2; P2: in out Skew2; UFac,VFac: in out Scalar);
  procedure RG3(P1: in Skew2; P2: in out Skew2);

  procedure DProdGFiG(DP: in Skew2; O: in PData; DH: in out Skew);
  procedure DProdGiFGiFGi(DP: in Skew2; O1,O2: in PData; DH: in out Skew);
  procedure DFactors(O: in RGData; DF: in Skew; DUFac,DVFac: in out Scalar);
  procedure DEqualize(O: in RGData; DP1: in Skew2; DP2: in out Skew2);
  procedure DRG3(O: in RGData; DP1: in Skew2; DP2: in out Skew2);

  type CMode is     --- a unit coeff or unit error, defines subspaces
    record          --- modes are needed to use the package Lin
      I: Power;     --- Taylor1 index
      J: Index4;    --- matrix element
      F: Boolean;   --- F component
      Err: Boolean; --- error bound
    end record;

  procedure Norm(M: in CMode; S: in out Scalar);
  procedure MaxNorm(M: in CMode; S: out Radius);
  procedure Assign(M: in CMode; P: in out Skew2);
  procedure AddProd(M: in CMode; S: in Scalar; P: in out Skew2; Check: in Boolean);
  procedure Extract(M: in CMode; P: in out Skew2; S: in out Scalar; Check: in Boolean);
  procedure Show(M: in CMode);
  function Reversible(M: CMode) return Boolean;

  type CModes is array (Positive range <>) of CMode; -- defined a "partition" of the space into a product

  procedure Show(M: in CModes);
  function Make(MaxDeg: Power; Trunc: Boolean) return CModes;

  --------------------------------------------------
  --- the contraction etc via generic procedures ---
  --------------------------------------------------

  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => Skew2, Mode => CMode, Modes => CModes);
  package Lin_Contr is new Lin.Contr (SSV => SSV);
  use Lin, Lin_Contr;

  QC: constant CModes := Make(Mat_PackSize,True);   --- no Err modes
  subtype QC_Vector is SVector(QC'Range);
  subtype QC_Matrix is SMatrix(QC'Range,QC'Range);

  procedure Extra_Info(P: in Skew2; UFac,VFac: in Scalar);
  procedure DContrNorm(M: in SMatrix; O: in RGData; PackSize: in Positive; R: out Radius);
  procedure DContrNorm(M: in SMatrix; P: in Skew2; PackSize: in Positive; R: out Radius);
  procedure ContrFix(M: in SMatrix; PackSize: in Positive; P: in out Skew2);

  function Fix_Name return String;
  function Eigen1_Name return String;
  function ContrMat_Name return String;
  function DerMatrix_Name return String;
  procedure NumTrunc(M: in out SMatrix; Cut: in Radius := Zero);

  ------- purely numerical
  procedure DerMatrix(O: in RGData; M: in out SMatrix);
  procedure DerMatrix(P: in Skew2; M: in out SMatrix);
  procedure ContrMatrix(O: in RGData; M: in out SMatrix);
  procedure ContrMatrix(P: in Skew2; M: in out SMatrix);
  procedure NumInterpolate(P0: in Skew2; P1: in out Skew2; Fac: in Flt);
  procedure NumIProd(P1,P2: in Skew2; S: in out Scalar);
  procedure AfterContr0(P: in out Skew2);
  procedure ApproxFix(P: in out Skew2; M: in out SMatrix; Eps: in Radius := Zero);
  procedure IterContr(P: in out Skew2; M: in out SMatrix);
  procedure ApproxNormalize(P: in out Skew2; Fac: in out Scalar);
  procedure ApproxEigen1(O: in RGData; H: in out Skew2; Eps: in Radius := Zero);
  procedure ApproxEigen1(P: in Skew2; H: in out Skew2; Eps: in Radius := Zero);
  procedure QC_UnPack(V: in QC_Vector; P: in out Skew2);

  DContr_Parallel: Boolean renames Lin.Op_Norm_Parallel;
  DerMatrix_Parallel: Boolean renames Lin.Enclosure_Parallel;

  --- rounded up
  InvRhoF:    constant Radius := One/RhoF;
  InvRhoG:    constant Radius := One/RhoG;
  RhoFPow:    constant RadVec := PowersOf(RhoF,PDeg);
  RhoGPow:    constant RadVec := PowersOf(RhoG,PDeg);
  InvRhoFPow: constant RadVec := PowersOf(InvRhoF,PDeg);
  InvRhoGPow: constant RadVec := PowersOf(InvRhoG,PDeg);

  IG1: constant Radius  := Flt(41475559)/Flt(67108864); --- upper bound on InvGold
  IG3: constant Radius  := Flt(7921127)/Flt(33554432); --- upper bound on InvGold**3
  IG2H: constant Radius := Flt(801041)/Flt(4194304);  --- upper bound on (InvGold**2)/2

  RhoHatF: constant Radius := IG3*RhoF;     -- domain for GFiG
  RhoHatG: constant Radius := IG3*RhoG;     -- domain for GiFGiFGi
  RhoFGiF: constant Radius := RhoHatG+IG2H; -- domain for FGiF

private

  procedure DEqualize(UFac,VFac,DUFac,DVFac: in Scalar; H1,DH1: in Skew; DH2: in out Skew);
  procedure Extract_Coeff(M: in CMode; P: in out Skew2; S: in out Scalar);
  procedure Extract_Err(M: in CMode; P: in out Skew2; S: in out Scalar);

  Tiny:     constant Radius := 1.0E-999;


end Taylors1.Skews.Pairs;
