with Globals, Protected_Counters, Strings, Flts.Plotting;
use Globals, Protected_Counters, Strings, Flts.Plotting;

pragma Elaborate_All (Globals,Protected_Counters,Strings,Flts.Plotting);

package body Taylors1.Skews is

  procedure Prod3(F1,G1: in Taylor1; H1: in out Taylor1;
                  F2,G2: in Taylor1; H2: in out Taylor1;
                  F3,G3: in Taylor1; H3: in out Taylor1) is
    ---
    Task_Error: Boolean := False;
    PC: Protected_Counter(1);

    procedure Prod_Loop is
      J: Positive;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        case J is
          when 1 => Prod(F1,G1,H1);
          when 2 => Prod(F2,G2,H2);
          when 3 => Prod(F3,G3,H3);
          when others => exit;
        end case;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(2,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Prod3;

  procedure Prod4(F1,G1: in Taylor1; H1: in out Taylor1;
                  F2,G2: in Taylor1; H2: in out Taylor1;
                  F3,G3: in Taylor1; H3: in out Taylor1;
                  F4,G4: in Taylor1; H4: in out Taylor1) is
    ---
    Task_Error: Boolean := False;
    PC: Protected_Counter(1);

    procedure Prod_Loop is
      J: Positive;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        case J is
          when 1 => Prod(F1,G1,H1);
          when 2 => Prod(F2,G2,H2);
          when 3 => Prod(F3,G3,H3);
          when 4 => Prod(F4,G4,H4);
          when others => exit;
        end case;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(3,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Prod4;

  procedure Prod8(F1,G1: in Taylor1; H1: in out Taylor1;
                  F2,G2: in Taylor1; H2: in out Taylor1;
                  F3,G3: in Taylor1; H3: in out Taylor1;
                  F4,G4: in Taylor1; H4: in out Taylor1;
                  F5,G5: in Taylor1; H5: in out Taylor1;
                  F6,G6: in Taylor1; H6: in out Taylor1;
                  F7,G7: in Taylor1; H7: in out Taylor1;
                  F8,G8: in Taylor1; H8: in out Taylor1) is
    ---
    Task_Error: Boolean := False;
    PC: Protected_Counter(1);

    procedure Prod_Loop is
      J: Positive;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        case J is
          when 1 => Prod(F1,G1,H1);
          when 2 => Prod(F2,G2,H2);
          when 3 => Prod(F3,G3,H3);
          when 4 => Prod(F4,G4,H4);
          when 5 => Prod(F5,G5,H5);
          when 6 => Prod(F6,G6,H6);
          when 7 => Prod(F7,G7,H7);
          when 8 => Prod(F8,G8,H8);
          when others => exit;
        end case;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(7,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Prod8;

  procedure Translate4(F1: in Taylor1; S1: in Scalar; G1: in out Taylor1;
                       F2: in Taylor1; S2: in Scalar; G2: in out Taylor1;
                       F3: in Taylor1; S3: in Scalar; G3: in out Taylor1;
                       F4: in Taylor1; S4: in Scalar; G4: in out Taylor1;
                       R: in Radius) is
    ---
    Task_Error: Boolean := False;
    PC: Protected_Counter(1);

    procedure Translate_Loop is
      J: Positive;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        case J is
          when 1 => Translate(F1,S1,G1,R);
          when 2 => Translate(F2,S2,G2,R);
          when 3 => Translate(F3,S3,G3,R);
          when 4 => Translate(F4,S4,G4,R);
          when others => exit;
        end case;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Translate_Loop;

    task type Translate_Task_Type is end Translate_Task_Type;
    task body Translate_Task_Type is begin Translate_Loop; end Translate_Task_Type;

    NT: constant Natural := Reserve_Tasks(3,Allow_Parallel);
  begin
    if NT=0 then
      Translate_Loop;
    else
      declare
        Translate_Task: array(1 .. NT) of Translate_Task_Type;
        pragma Warnings (Off,Translate_Task);
      begin
        Translate_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Translate4;

  ---------------------------------------
  ---------------------------------------
  ---------------------------------------

  procedure Inv(M: in out SMat2) is
    --- ABCD representation
    --- modulo a factor 1/det(M)
  begin
    Neg(M(JB));
    Neg(M(JC));
    Swap(M(JA),M(JD));
  end Inv;

  procedure TUVS2ABCD(M: in out SMat2) is
    A,D: Scalar;
  begin
    Sum(M(JT),M(JS),A);
    Diff(M(JT),M(JS),D);
    Swap(A,M(JA));
    Swap(D,M(JD));
  end TUVS2ABCD;

  procedure TUVS2ABCD(F: in out TMat2; Fix_Rev: in Boolean := False) is
    Tmp1,Tmp2: LT_Pointer;
  begin
    if Fix_Rev then Fix_Reversible(F); end if;
    TPool.Allocate(Tmp1);
    TPool.Allocate(Tmp2);
    declare
      A: Taylor1 renames Tmp1.all.Data;
      D: Taylor1 renames Tmp2.all.Data;
    begin
      Sum(F(JT),F(JS),A);
      Diff(F(JT),F(JS),D);
      Swap(A,F(JA));
      Swap(D,F(JD));
    end;
    TPool.Recycle(Tmp1);
    TPool.Recycle(Tmp2);
  end TUVS2ABCD;

  procedure TUVS2ABCD(F: in TMat2; G: in out TMat2; Fix_Rev: in Boolean := False) is
    T: Taylor1 renames G(JB);
    S: Taylor1 renames G(JC);
  begin
    Copy(F(JT),T);
    Copy(F(JS),S);
    if Fix_Rev then
      CutParity(T,1);
      CutParity(S,0);
    end if;
    Sum(T,S,G(JA));
    Diff(T,S,G(JD));
    Copy(F(JU),G(JB));
    Copy(F(JV),G(JC));
    if Fix_Rev then
      CutParity(G(JB),1);
      CutParity(G(JC),1);
    end if;
  end TUVS2ABCD;

  procedure ABCD2TUVS(F: in out TMat2; Fix_Rev: in Boolean := False) is
    Tmp1,Tmp2: LT_Pointer;
  begin
    TPool.Allocate(Tmp1);
    TPool.Allocate(Tmp2);
    declare
      T: Taylor1 renames Tmp1.all.Data;
      S: Taylor1 renames Tmp2.all.Data;
    begin
      Sum(F(JA),F(JD),T);
      Diff(F(JA),F(JD),S);
      Prod(SHalf,T,F(JT));
      Prod(SHalf,S,F(JS));
    end;
    TPool.Recycle(Tmp1);
    TPool.Recycle(Tmp2);
    if Fix_Rev then Fix_Reversible(F); end if;
  end ABCD2TUVS;

  procedure ABCD2TUVS(F: in TMat2; G: in out TMat2; Fix_Rev: in Boolean := False) is
    A: Taylor1 renames G(JU);
    D: Taylor1 renames G(JV);
  begin
    Prod(SHalf,F(JA),A);
    Prod(SHalf,F(JD),D);
    Sum(A,D,G(JT));
    Diff(A,D,G(JS));
    Copy(F(JB),G(JU));
    Copy(F(JC),G(JV));
    if Fix_Rev then Fix_Reversible(G); end if;
  end ABCD2TUVS;

  procedure Copy(F: in TMat2; G: in out TMat2) is
  begin
    for J in Index4 loop Copy(F(J),G(J)); end loop;
  end Copy;

  procedure Add(F: in TMat2; G: in out TMat2) is
  begin
    for J in Index4 loop Add(F(J),G(J)); end loop;
  end Add;

  procedure Det(F: in TMat2; P0: in out Taylor1) is
    --- P0:= T*T-S*S-U*V
    Tmp1,Tmp2: LT_Pointer;
  begin
    TPool.Allocate(Tmp1);
    TPool.Allocate(Tmp2);
    declare
      P1: Taylor1 renames Tmp1.all.Data;
      P2: Taylor1 renames Tmp2.all.Data;
    begin
      Prod3(F(JT),F(JT),P0,  -- T*T
            F(JS),F(JS),P1,  -- S*S
            F(JU),F(JV),P2); -- U*V
      Sub(P1,P0);
      Sub(P2,P0);
    end;
    TPool.Recycle(Tmp1);
    TPool.Recycle(Tmp2);
  end Det;

  procedure ShowEigenVal(N: in String; M: in SMat2; Det1: in Boolean := False) is
    --- M in ABCD format
    B,C,U1,U2,V: Scalar;
  begin
    Sum(M(JA),M(JD),B);
    Neg(B);                -- B := -trace
    if Det1 then
      Copy(SOne,C);
    else
      Prod(M(JA),M(JD),C);
      Prod(M(JB),M(JD),V);
      Sub(V,C);            -- C := det
    end if;
    Roots2(B,C,U1,U2,V);
    Show2(N & "1st eigen:",U1,V);
    Neg(V);
    Show2(N & "2nd eigen:",U2,V);
  end ShowEigenVal;

  procedure DDet(F,DF: in TMat2; P0: in out Taylor1) is
    --- P0 := 2*T*DT-2*S*DS-U*DV-V*DU
    Tmp1,Tmp2,Tmp3: LT_Pointer;
  begin
    TPool.Allocate(Tmp1);
    TPool.Allocate(Tmp2);
    TPool.Allocate(Tmp3);
    declare
      P1: Taylor1 renames Tmp1.all.Data;
      P2: Taylor1 renames Tmp2.all.Data;
      P3: Taylor1 renames Tmp3.all.Data;
    begin
      Prod4(F(JT),DF(JT),P0,  -- T*DT
            F(JS),DF(JS),P1,  -- S*DS
            F(JU),DF(JV),P2,  -- U*DV
            F(JV),DF(JU),P3); -- V*DU
      Sub(P1,P0);
      Mult(STwo,P0);
      Sub(P2,P0);
      Sub(P3,P0);
    end;
    TPool.Recycle(Tmp1);
    TPool.Recycle(Tmp2);
    TPool.Recycle(Tmp3);
  end DDet;

  procedure Prod(M1,M2: in SMat2; M3: in out SMat2) is
  begin
    Prod(M1(J11),M2(J11),M3(J11));
    AddProd(M1(J12),M2(J21),M3(J11));
    Prod(M1(J11),M2(J12),M3(J12));
    AddProd(M1(J12),M2(J22),M3(J12));
    Prod(M1(J21),M2(J11),M3(J21));
    AddProd(M1(J22),M2(J21),M3(J21));
    Prod(M1(J21),M2(J12),M3(J22));
    AddProd(M1(J22),M2(J22),M3(J22));
  end Prod;

  procedure Prod(M: in SMat2; F: in TMat2; G: in out TMat2) is
  begin
    Prod(M(J11),F(J11),G(J11));
    AddProd(M(J12),F(J21),G(J11));
    Prod(M(J11),F(J12),G(J12));
    AddProd(M(J12),F(J22),G(J12));
    Prod(M(J21),F(J11),G(J21));
    AddProd(M(J22),F(J21),G(J21));
    Prod(M(J21),F(J12),G(J22));
    AddProd(M(J22),F(J22),G(J22));
  end Prod;

  procedure Prod(F: in TMat2; M: in SMat2; G: in out TMat2) is
  begin
    Prod(M(J11),F(J11),G(J11));
    AddProd(M(J21),F(J12),G(J11));
    Prod(M(J12),F(J11),G(J12));
    AddProd(M(J22),F(J12),G(J12));
    Prod(M(J11),F(J21),G(J21));
    AddProd(M(J21),F(J22),G(J21));
    Prod(M(J12),F(J21),G(J22));
    AddProd(M(J22),F(J22),G(J22));
  end Prod;

  procedure Prod(P: in Taylor1; F: in TMat2; G: in out TMat2) is
  begin
    Prod4(P,F(JT),G(JT),
          P,F(JU),G(JU),
          P,F(JV),G(JV),
          P,F(JS),G(JS));
  end Prod;

  procedure Prod(F,G: in TMat2; H: in out TMat2) is
    FF,GG,HH: Skew;
    pragma Warnings (Off,FF);
    pragma Warnings (Off,GG);
    pragma Warnings (Off,HH);
    Ft: TMat2 renames FF.L.all.Data;
    Gt: TMat2 renames GG.L.all.Data;
    Ht: TMat2 renames HH.L.all.Data;
  begin
    TUVS2ABCD(F,Ft);
    TUVS2ABCD(G,Gt);
    Prod8(Ft(J11),Gt(J11),Ht(J11),
          Ft(J12),Gt(J21),H(J11),
          --
          Ft(J11),Gt(J12),Ht(J12),
          Ft(J12),Gt(J22),H(J12),
          ---
          Ft(J21),Gt(J11),Ht(J21),
          Ft(J22),Gt(J21),H(J21),
          ----
          Ft(J21),Gt(J12),Ht(J22),
          Ft(J22),Gt(J22),H(J22));
    --
    for J in Index4 loop Add(H(J),Ht(J)); end loop;
    ABCD2TUVS(Ht,H);
  end Prod;

  procedure Translate(F: in TMat2; S: in Scalar; G: in out TMat2; R: in Radius) is
  begin
    Translate4(F(JT),S,G(JT),
               F(JU),S,G(JU),
               F(JV),S,G(JV),
               F(JS),S,G(JS),R);
  end Translate;

  procedure Conjugate(F: in out TMat2) is
    --- a NOP if F is reversible
  begin
    Reflect(F(JT),1);
    Reflect(F(JU),1);
    Reflect(F(JV),1);
    Reflect(F(JS),0);
  end Conjugate;

  procedure InvConjugate(F: in out TMat2) is
    --- first Conjugate, then Inv
  begin
    Reflect(F(JT),1);
    Reflect(F(JU),0);
    Reflect(F(JV),0);
    Reflect(F(JS),1);
  end InvConjugate;

  procedure Fix_Reversible(F: in out TMat2) is
  begin
    CutParity(F(JT),1);
    CutParity(F(JU),1);
    CutParity(F(JV),1);
    CutParity(F(JS),0);
  end Fix_Reversible;

  procedure Factors(M: in TMat2; UFac,VFac: in out Scalar) is
    --- factors used in Equalize
    UX,VX,Fac: Scalar;
  begin
    Val(M(JU),SZero,UX);
    Val(M(JV),SZero,VX);
    ---
    Quot(VX,UX,Fac);
    if Sign(Fac)<0 then Neg(Fac); end if;
    Sqrt(Fac,UFac);
    Quot(UX,VX,Fac);
    if Sign(Fac)<0 then Neg(Fac); end if;
    Sqrt(Fac,VFac);
    ---
    if Sign(UX)<0 then
      Neg(UFac);
      Neg(VFac);
    end if;
  end Factors;

  ----------------------------------------
  ----------------------------------------
  ----------------------------------------

  function CName(J: Index4) return Character is
  begin
    case J is
      when JT => return 'T';
      when JU => return 'U';
      when JV => return 'V';
      when JS => return 'S';
    end case;
  end CName;

  function IsZero(F: Skew) return Boolean is
    M: TMat2 renames F.L.all.Data;
  begin
    for J in Index4 loop
      if not IsZero(M(J)) then return False; end if;
    end loop;
    return True;
  end IsZero;

  procedure SetZero(R: in Radius; F: in out Skew) is
    M: TMat2 renames F.L.all.Data;
  begin
    F.Det1 := False;
    F.Alpha := GZero;
    for J in Index4 loop SetZero(R,M(J)); end loop;
  end SetZero;

  procedure Copy(F1: in Skew; F2: in out Skew) is
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
  begin
    F2.Det1 := F1.Det1;
    F2.Alpha := F1.Alpha;
    for J in Index4 loop Copy(M1(J),M2(J)); end loop;
  end Copy;

  procedure Swap(F1,F2: in out Skew) is
    use TP.Alloc;
    Det12: constant Boolean := F2.Det1;
    Alpha2: constant GNum := F2.Alpha;
    L2: constant LD_Pointer := F2.L;
  begin
    F2.Det1 := F1.Det1; F1.Det1 := Det12;
    F2.Alpha := F1.Alpha; F1.Alpha := Alpha2;
    F2.L := F1.L; F1.L := L2;
  end Swap;

  procedure Neg(F: in out Skew) is
    M: TMat2 renames F.L.all.Data;
  begin
    for J in Index4 loop Neg(M(J)); end loop;
  end Neg;

  procedure Neg(F1: in Skew; F2: in out Skew) is
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
  begin
    F2.Det1 := F1.Det1;
    F2.Alpha := F1.Alpha;
    for J in Index4 loop Neg(M1(J),M2(J)); end loop;
  end Neg;

  procedure Add(F1: in Skew; F2: in out Skew) is
  begin
    if F1.Alpha=F2.Alpha then
      F2.Det1 := False;
      declare
        M1: TMat2 renames F1.L.all.Data;
        M2: TMat2 renames F2.L.all.Data;
      begin
        for J in Index4 loop Add(M1(J),M2(J)); end loop;
      end;
    elsif IsZero(F1) then
      null;
    elsif IsZero(F2) then
      Copy(F1,F2);
    else
      Show2("foooo ",F1,F2);
      raise Undefined;
    end if;
  end Add;

  procedure Sum(F1,F2: in Skew; F3: in out Skew) is
  begin
    if F1.Alpha=F2.Alpha then
      F3.Det1 := False;
      F3.Alpha := F1.Alpha;
      declare
        M1: TMat2 renames F1.L.all.Data;
        M2: TMat2 renames F2.L.all.Data;
        M3: TMat2 renames F3.L.all.Data;
      begin
        for J in Index4 loop Sum(M1(J),M2(J),M3(J)); end loop;
      end;
    elsif IsZero(F1) then
      Copy(F2,F3);
    elsif IsZero(F2) then
      Copy(F1,F3);
    else
      Show2("foooo ",F1,F2);
      raise Undefined;
    end if;
  end Sum;

  procedure Sub(F1: in Skew; F2: in out Skew) is
  begin
    if F1.Alpha=F2.Alpha then
      F2.Det1 := False;
      declare
        M1: TMat2 renames F1.L.all.Data;
        M2: TMat2 renames F2.L.all.Data;
      begin
        for J in Index4 loop Sub(M1(J),M2(J)); end loop;
      end;
    elsif IsZero(F1) then
      null;
    elsif IsZero(F2) then
      Neg(F1,F2);
    else
      Show2("foooo ",F1,F2);
      raise Undefined;
    end if;
  end Sub;

  procedure Diff(F1,F2: in Skew; F3: in out Skew) is
  begin
    if F1.Alpha=F2.Alpha then
      F3.Det1 := False;
      F3.Alpha := F1.Alpha;
      declare
        M1: TMat2 renames F1.L.all.Data;
        M2: TMat2 renames F2.L.all.Data;
        M3: TMat2 renames F3.L.all.Data;
      begin
        for J in Index4 loop Diff(M1(J),M2(J),M3(J)); end loop;
      end;
    elsif IsZero(F1) then
      Neg(F2,F3);
    elsif IsZero(F2) then
      Copy(F1,F3);
    else
      Show2("foooo ",F1,F2);
      raise Undefined;
    end if;
  end Diff;

  procedure Val(F: in Skew; S: in Scalar; FS: in out SMat2) is
    M: TMat2 renames F.L.all.Data;
  begin
    for J in Index4 loop Val(M(J),S,FS(J)); end loop;
  end Val;

  procedure Component(F: in Skew; I: in Power; J: in Index4; S: in out Scalar) is
    M: TMat2 renames F.L.all.Data;
  begin
    Component(I,M(J),S);
  end Component;

  procedure TraceAt0(F: in Skew; S: in out Scalar) is
    M: TMat2 renames F.L.all.Data;
  begin
    Component(0,M(JT),S);
    Mult(STwo,S);
  end TraceAt0;

  procedure SameRho(F: in out Skew) is
    -- avoid that zero components have a different Rho
    M: TMat2 renames F.L.all.Data;
    MaxR: Radius := Zero;
    MinR: Radius := Radius'Last;
  begin
    for J in Index4 loop
      if M(J).R>MaxR or M(J).R<MinR then
        if not IsZero(M(J)) then
          if M(J).R>MaxR then MaxR := M(J).R; end if;
          if M(J).R<MinR then MinR := M(J).R; end if;
        end if;
      end if;
    end loop;
    if MaxR>MinR then raise Sorry; end if;
    if MaxR<MinR then MinR := BigRho; end if;      -- F=0
    for J in Index4 loop M(J).R := MinR; end loop;
  end SameRho;

  function WarnRho(N: String; F: Skew) return Radius is
    M: TMat2 renames F.L.all.Data;
    MaxR: Radius := Zero;
    MinR: Radius := Radius'Last;
  begin
    for J in Index4 loop
      if M(J).R>MaxR or M(J).R<MinR then
        if not IsZero(M(J)) then
          if M(J).R>MaxR then MaxR := M(J).R; end if;
          if M(J).R<MinR then MinR := M(J).R; end if;
        end if;
      end if;
    end loop;
    if MaxR<MinR then return BigRho; end if; -- F=0
    if MaxR>MinR then
      Show1(N & "Warning: T.R =",M(JT).R);
      Show1(N & "Warning: U.R =",M(JU).R);
      Show1(N & "Warning: V.R =",M(JV).R);
      Show1(N & "Warning: S.R =",M(JS).R);
    end if;
    return MinR;
  end WarnRho;

  function Rho(F: Skew) return Radius is
    M: TMat2 renames F.L.all.Data;
    R: Radius := BigRho;
  begin
    for J in Index4 loop
      if M(J).R<R then
        if not IsZero(M(J)) then R := M(J).R; end if;
      end if;
    end loop;
    return R;
  end Rho;

  procedure SetRho(R: in Radius; F: in out Skew; Check: in Boolean := True) is
    M: TMat2 renames F.L.all.Data;
  begin
    for J in Index4 loop SetRho(R,M(J),Check); end loop;
  end SetRho;

  procedure Norm(R: in Radius; F: in Skew; S: in out Scalar) is
    M: TMat2 renames F.L.all.Data;
    N: Scalar;
  begin
    SetZero(S);
    for J in Index4 loop
      Norm1(R,M(J),N);
      Add(N,S);
    end loop;
  end Norm;

  procedure Norm(F: in Skew; S: in out Scalar) is
  begin
    Norm(Rho(F),F,S);
  end Norm;

  function MaxNorm(R: Radius; F: Skew) return Radius is
    M: TMat2 renames F.L.all.Data;
    S: Radius := Zero;
  begin
    for J in Index4 loop
      S := S+MaxNorm(R,M(J));
    end loop;
    return S;
  end MaxNorm;

  function MaxNorm(F: Skew) return Radius is
  begin
    return MaxNorm(Rho(F),F);
  end MaxNorm;

  function MaxDist(R: Radius; F1,F2: Skew) return Radius is
    F3: Skew;
  begin
    Diff(F1,F2,F3);
    return MaxNorm(R,F3);
  end MaxDist;

  function MaxDist(F1,F2: Skew) return Radius is
    F3: Skew;
  begin
    Diff(F1,F2,F3);
    return MaxNorm(F3);
  end MaxDist;

  ---------------------------------
  ---------------------------------
  ---------------------------------

  procedure Inv(F: in out Skew) is
    M: TMat2 renames F.L.all.Data;
  begin
    F.Alpha := -F.Alpha;
    Neg(M(JU));
    Neg(M(JV));
    Neg(M(JS));
  end Inv;

  procedure Inv(F1: in Skew; F2: in out Skew) is
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
  begin
    F2.Det1 := F1.Det1;
    F2.Alpha := -F1.Alpha;
    Copy(M1(JT),M2(JT));
    Neg(M1(JU),M2(JU));
    Neg(M1(JV),M2(JV));
    Neg(M1(JS),M2(JS));
  end Inv;

  procedure Prod(S: in Scalar; F1: in Skew; F2: in out Skew) is
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
  begin
    F2.Det1 := False;
    F2.Alpha := F1.Alpha;
    for J in Index4 loop Prod(S,M1(J),M2(J)); end loop;
  end Prod;

  procedure Mult(S: in Scalar; F: in out Skew) is
    M: TMat2 renames F.L.all.Data;
  begin
    F.Det1 := False;
    for J in Index4 loop Mult(S,M(J)); end loop;
  end Mult;

  procedure Prod(P: in Taylor1; F1: in Skew; F2: in out Skew) is
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
  begin
    F2.Det1 := False;
    F2.Alpha := F1.Alpha;
    Prod(P,M1,M2);
  end Prod;

  procedure Mult(P: in Taylor1; F: in out Skew) is
    Tmp: Skew;
  begin
    Prod(P,F,Tmp);
    Swap(Tmp,F);
  end Mult;

  procedure AddProd(P: in Taylor1; F1: in Skew; F2: in out Skew) is
    Tmp: Skew;
  begin
    Prod(P,F1,Tmp);
    Add(Tmp,F2);
  end AddProd;

  procedure Prod(F1,F2: in Skew; F3: in out Skew; R: in Radius) is
    --- the two Translate could be parallelized
    S,S1,S2: Scalar;
    Ft1,Ft2: Skew;
    pragma Warnings (Off,Ft1);
    pragma Warnings (Off,Ft2);
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
    M3: TMat2 renames F3.L.all.Data;
    Mt1: TMat2 renames Ft1.L.all.Data; -- uses Ft1
    Mt2: TMat2 renames Ft2.L.all.Data; -- uses Ft2
  begin
    F3.Det1 := F1.Det1 and F2.Det1;
    F3.Alpha := F1.Alpha+F2.Alpha;
    Assign(F2.Alpha,S);
    Prod(SHalf,S,S1);
    Translate(M1,S1,Mt1,R);
    Assign(F1.Alpha,S);
    Prod(SNegHalf,S,S2);
    Translate(M2,S2,Mt2,R);
    Prod(Mt1,Mt2,M3);
  end Prod;

  procedure DProd(F,G,DF,DG: in Skew; DH: in out Skew; R: in Radius) is
    FDG: Skew;
  begin
    Prod(F,DG,FDG,R);
    Prod(DF,G,DH,R);
    Add(FDG,DH);      --- DH := F*DG+DF*G
  end DProd;

  procedure Reflect(F: in out Skew; Offset: in Parity) is
    --- time reflection of M if Offset=1
    M: TMat2 renames F.L.all.Data;
  begin
    for J in Index4 loop Reflect(M(J),Offset); end loop;
  end Reflect;

  procedure Scale1(F: in out Skew) is
    --- scale by Inv_Gold
    M: TMat2 renames F.L.all.Data;
    S: Scalar;
  begin
    F.Alpha := Gold*F.Alpha;
    Assign(Inv_Gold,S);
    for J in Index4 loop Scale(S,M(J)); end loop;
  end Scale1;

  procedure Scale3(F: in out Skew) is
    --- scale by Inv_Gold**3
    M: TMat2 renames F.L.all.Data;
    S: Scalar;
  begin
    F.Alpha := (Gold**3)*F.Alpha;
    Assign(Inv_Gold**3,S);
    for J in Index4 loop Scale(S,M(J)); end loop;
  end Scale3;

  procedure Translate(F: in Skew; S: in Scalar; G: in out Skew; R: in Radius) is
    MF: TMat2 renames F.L.all.Data;
    MG: TMat2 renames G.L.all.Data;
  begin
    if IsZero(S) then
      Copy(F,G);
    else
      G.Det1 := F.Det1;
      G.Alpha := F.Alpha;
      Translate(MF,S,MG,R);
    end if;
  end Translate;

  function DetMinus1(F: Skew; R: Radius) return Radius is
    M: TMat2 renames F.L.all.Data;
    E: Radius;
    Tmp: LT_Pointer;
  begin
    TPool.Allocate(Tmp);
    declare
      P: Taylor1 renames Tmp.all.Data;
    begin
      Det(M,P);
      Sub(SOne,P.C(0));
      E := MaxNorm(R,P);
    end;
    TPool.Recycle(Tmp);
    return E;
  end DetMinus1;

  procedure Show_Determinant(N: in String; F: in Skew) is
    M: TMat2 renames F.L.all.Data;
    Tmp: LT_Pointer;
  begin
    TPool.Allocate(Tmp);
    Det(M,Tmp.all.Data);
    Show1(N & "det ",Tmp.all.Data);
    TPool.Recycle(Tmp);
  end Show_Determinant;

  procedure Fix_Rho(F: in out Skew) is
    R: constant Radius:= Rho(F);
    M: TMat2 renames F.L.all.Data;
  begin
    for J in Index4 loop M(J).R := R; end loop;
  end Fix_Rho;

  procedure Fix_Reversible(F: in out Skew) is
    M: TMat2 renames F.L.all.Data;
  begin
    CutParity(M(JT),1);
    CutParity(M(JU),1);
    CutParity(M(JV),1);
    CutParity(M(JS),0);
  end Fix_Reversible;

  procedure Show_Reversible(N: in String; F: in Skew) is
    M: TMat2 renames F.L.all.Data;
  begin
    Show1(N & "T Rev err",ParityErr(M(JT),1));
    Show1(N & "U Rev err",ParityErr(M(JU),1));
    Show1(N & "V Rev err",ParityErr(M(JV),1));
    Show1(N & "S Rev err",ParityErr(M(JS),0));
  end Show_Reversible;

  procedure AddBall(R: in Radius; S: in Scalar; F: in out Skew) is
    --- improve this
    B: constant Radius := MaxNorm(S);
    M: TMat2 renames F.L.all.Data;
  begin
    if Not_Strunc then
      for J in Index4 loop
        declare
          P: Taylor1 renames M(J);
        begin
          if P.R<R and then P.F<PDeg1 then raise Domain_Error; end if;
          P.R := R;
          if J=JS then
            Add_Ball(1,B,P); --- also sets P.F := 1
          else
            Add_Ball(0,B,P); --- also sets P.F := 0
          end if;
        end;
      end loop;
    end if;
  end AddBall;

  procedure NumTrunc(F: in out Skew; Deg: in Power := PDeg; Cut: in Radius := Zero) is
    M: TMat2 renames F.L.all.Data;
  begin
    for J in Index4 loop NumTrunc(M(J),Deg,Cut); end loop;
  end NumTrunc;

  procedure NumInterpolate(F0: in Skew; F1: in out Skew; Fac: in Flt) is
    --- purely numerical
    --- F1 := (1-Fac)*F0+Fac*F1
    M0: TMat2 renames F0.L.all.Data;
    M1: TMat2 renames F1.L.all.Data;
  begin
    if Fac /= One then
      for J in Index4 loop NumInterpolate(M0(J),M1(J),Fac); end loop;
    end if;
  end NumInterpolate;

  procedure NumIProd(F1,F2: in Skew; S: in out Scalar) is
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
    N: Scalar;
  begin
    SetZero(S);
    for J in Index4 loop
      NumIProd(M1(J),M2(J),N);
      Add(N,S);
    end loop;
  end NumIProd;

  function GuessRho(F: Skew) return Radius is
    M: TMat2 renames F.L.all.Data;
    R: Radius := Zero;
  begin
    for J in Index4 loop
      R := R+GuessRho(M(J));
    end loop;
    return Quarter*R;
  end GuessRho;

  ---------------------------------
  ---------------------------------
  ---------------------------------

  function Strng(N: GNum) return String is
  begin
    return "(" & Strng(N.I) & "," & Strng(N.G) & ")";
  end Strng;

  procedure ShowType1(N: in String; F: in Skew) is
    R: constant Radius := WarnRho(N,F);
  begin
    Show1(N & "Det1 ",F.Det1);
    Show0(N & "Alpha " & Strng(F.Alpha));
    Show1(N & "R",R);
  end ShowType1;

  procedure ShowType2(N: in String; F1,F2: in Skew) is
    R1: constant Radius := WarnRho(N & "1 ",F1);
    R2: constant Radius := WarnRho(N & "2 ",F2);
  begin
    Show2(N & "Det1 ",F1.Det1,F2.Det1);
    Show0(N & "Alpha " & Strng(F1.Alpha) & " " & Strng(F2.Alpha));
    Show2(N & "R",R1,R2);
  end ShowType2;

  procedure ShowNorms(N: in String; F: in Skew; ShowType: in Boolean := True) is
    M: TMat2 renames F.L.all.Data;
    R: constant Radius := WarnRho(N,F);
  begin
    if ShowType then ShowType1(N,F); end if;
    for J in Index4 loop
      Show1(N & CName(J),MaxNorm(R,M(J)));
    end loop;
  end ShowNorms;

  procedure Show1(N: in String; F: in Skew) is
    Hide0: constant Boolean := True;
    M: TMat2 renames F.L.all.Data;
  begin
    ShowType1(N,F);
    for J in Index4 loop
      Show1(N & CName(J) & " ",M(J),Hide0);
    end loop;
  end Show1;

  procedure Show2(N: in String; F1,F2: in Skew) is
    Hide0: constant Boolean := True;
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
  begin
    ShowType2(N,F1,F2);
    for J in Index4 loop
      Show2(N & CName(J) & " ",M1(J),M2(J),Hide0);
    end loop;
  end Show2;

  procedure Put(F: in File_Type; G: in Skew; Decimal: in Boolean := False) is
    M: TMat2 renames G.L.all.Data;
  begin
    if G.Det1 then Put(F,1); else Put(F,0); end if;
    Put(F,G.Alpha.I);
    Put(F,G.Alpha.G);
    Put(F,M(JT),Decimal);
    Put(F,M(JU),Decimal);
    Put(F,M(JV),Decimal);
    Put(F,M(JS),Decimal);
  end Put;

  procedure Get(F: in File_Type; G: in out Skew; Decimal: in Boolean := False) is
    Det1: Parity := 0;
    M: TMat2 renames G.L.all.Data;
  begin
    Get(F,Det1);
    G.Det1 := (Det1=1);
    Get(F,G.Alpha.I);
    Get(F,G.Alpha.G);
    Get(F,M(JT),Decimal);
    Get(F,M(JU),Decimal);
    Get(F,M(JV),Decimal);
    Get(F,M(JS),Decimal);
  end Get;

  procedure Write(FileName: in String; G: in Skew; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,G,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; G: in out Skew; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,G,Decimal);
    Close(F);
  end Read;

  procedure PutPlot(F: in File_Type; P: in Taylor1; XL,XR: in Flt) is
    N: constant Positive := 256;
    DX: constant Flt := (XR-XL)/Flt(N);
    C: Poly1 renames P.C;
    X: Flt := XL;
    S,T: Scalar;
    use SP;
  begin
    for K in 0 .. N loop
      Assign(X,S);
      Evaluate(C,S,T);
      Plot2(F,X,Approx(T));
      X := X+DX;
    end loop;
    New_Line(F);
  end PutPlot;

  procedure WritePlot(FileName: in String; P: in Taylor1; XL,XR: in Flt) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    PutPlot(F,P,XL,XR);
    Close(F);
  end WritePlot;

  procedure WritePlots(Prefix: in String; F: in Skew; XL,XR: in Flt) is
    M: TMat2 renames F.L.all.Data;
  begin
    WritePlot(Prefix & "-t.plot",M(JT),XL,XR);
    WritePlot(Prefix & "-u.plot",M(JU),XL,XR);
    WritePlot(Prefix & "-v.plot",M(JV),XL,XR);
    WritePlot(Prefix & "-s.plot",M(JS),XL,XR);
  end WritePlots;

  ---------------------------------
  ---------------------------------
  ---------------------------------

  procedure FZero(F: in out Skew; R: in Radius) is
    --- translation by 1
    M: TMat2 renames F.L.all.Data;
  begin
    F.Det1 := True;
    F.Alpha := GOne;
    Assign(One,M(JT));
    SetZero(M(JU));
    SetZero(M(JV));
    SetZero(M(JS));
    for J in Index4 loop M(J).R := R; end loop;
  end FZero;

  procedure RevCosAM(E: in Scalar; F: in out Skew; R: in Radius) is
    --- reversible Almost Mathieu
    M: TMat2 renames F.L.all.Data;
    K: Scalar;
  begin
    F.Det1 := True;
    F.Alpha := Inv_Gold;
    ArcCos(SZero,K); -- Pi/2
    Mult(Four,K);    -- 2*Pi
    CosTaylor(K,SZero,R,M(JT));
    Neg(M(JT));
    Prod(SHalf,E,K);
    AddCoeff(0,K,M(JT));       -- T := E/2-Cos(2*Pi*.)
    SetZero(M(JS));            -- S := 0
    Copy(M(JT),M(JU));
    AddCoeff(0,SOne,M(JU));    -- U := T+1
    Copy(M(JT),M(JV));
    AddCoeff(0,SNegOne,M(JV)); -- V := T-1
    for J in Index4 loop M(J).R := R; end loop;
  end RevCosAM;

  procedure Equalize(F,G: in out Skew; UFac,VFac: in out Scalar) is
    --- make U=|V|
    MF: TMat2 renames F.L.all.Data;
    MG: TMat2 renames G.L.all.Data;
  begin
    Factors(MF,UFac,VFac);
    Mult(UFac,MF(JU));
    Mult(VFac,MF(JV));
    Mult(UFac,MG(JU));
    Mult(VFac,MG(JV));
  end Equalize;

  procedure Normalize(F1: in Skew; Fac1: in Taylor1; F2: in out Skew) is
    --- make determinant 1 (with proper Fac1)
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
  begin
    F2.Det1 := True;
    F2.Alpha := F1.Alpha;
    if F1.Det1 then
      Copy(M1,M2);
    else
      Prod(Fac1,M1,M2);
    end if;
  end Normalize;

  procedure Normalize(F1: in Skew; F2: in out Skew) is
    --- make determinant 1
    M1: TMat2 renames F1.L.all.Data;
    M2: TMat2 renames F2.L.all.Data;
    Tmp1: LT_Pointer;
    Tmp2: LT_Pointer;
  begin
    F2.Det1 := True;
    F2.Alpha := F1.Alpha;
    if F1.Det1 then
      Copy(M1,M2);
    else
      TPool.Allocate(Tmp1);
      TPool.Allocate(Tmp2);
      declare
        DetM: Taylor1 renames Tmp1.all.Data;
        Fac1: Taylor1 renames Tmp2.all.Data;
      begin
        Det(M1,DetM);
        QPower(IRat(-1,2),DetM,Fac1);   -- Fac1 := det(M)^{-1/2}
        Prod(Fac1,M1,M2);
      end;
      TPool.Recycle(Tmp1);
      TPool.Recycle(Tmp2);
    end if;
  end Normalize;

  procedure DetFactors(F: in Skew; Fac1,Fac3: in out Taylor1) is
    --- the two QPower could be parallelized
    Tmp: LT_Pointer;
    R: Radius;
  begin
    if F.Det1 then
      R := Rho(F);
      SetOne(R,Fac1);
      SetOne(R,Fac3);
    else
      TPool.Allocate(Tmp);
      declare
        M: TMat2 renames F.L.all.Data;
        DetM: Taylor1 renames Tmp.all.Data;
      begin
        Det(M,DetM);
        QPower(IRat(-1,2),DetM,Fac1); -- Fac1 := det(M)^{-1/2}
        QPower(IRat(-3,2),DetM,Fac3); -- Fac3 := det(M)^{-3/2}
      end;
      TPool.Recycle(Tmp);
    end if;
  end DetFactors;

  procedure DNormalize(F,DF: in Skew; Fac1,Fac3: in Taylor1; H: in out Skew) is
    M: TMat2 renames F.L.all.Data;
    DM: TMat2 renames DF.L.all.Data;
    Tmp: LT_Pointer;
  begin
    Copy(DF,H);                -- H := DF
    TPool.Allocate(Tmp);
    declare
      P: Taylor1 renames Tmp.all.Data;
    begin
      DDet(M,DM,P);
      Mult(SNegHalf,P);        -- P := -[2*T*DT-2*S*DS-U*DV-V*DU]/2
      if not F.Det1 then
        Mult(Fac1,H);          -- H := Fac1*DF
        Mult(Fac3,P);          -- P := Fac3*P
      end if;
      AddProd(P,F,H);
    end;
    TPool.Recycle(Tmp);
  end DNormalize;

  procedure Pre_Prod_GFG(F,G: in Skew; O: in out PData) is
  begin
    Copy(F,O.F);
    Copy(G,O.G);
    O.G_Right.Det1 := G.Det1;
    O.G_Left.Det1 := G.Det1;
    O.FG_Right.Det1 := F.Det1 and G.Det1;
    O.G_Right.Alpha := G.Alpha;
    O.G_Left.Alpha := G.Alpha;
    O.FG_Right.Alpha := F.Alpha+G.Alpha;
  end Pre_Prod_GFG;

  procedure Prod_GFG(F,G: in Skew; O: in out PData; H: in out Skew; R: in Radius) is
    --- H := GFG if F and G are reversible
    --- here Right and Left refer to translations
    B: TMat2 renames F.L.all.Data;
    A: TMat2 renames G.L.all.Data;
    C: TMat2 renames H.L.all.Data;
    ---
    A_Left: TMat2 renames O.G_Left.L.all.Data;
    A_Right: TMat2 renames O.G_Right.L.all.Data;
    BA_Right: TMat2 renames O.FG_Right.L.all.Data;
    S: Scalar;
  begin
    O.R := R;
    Pre_Prod_GFG(F,G,O);
    H.Det1 := F.Det1 and G.Det1;
    H.Alpha := F.Alpha+2*G.Alpha;
    ---
    Assign(O.FG_Right.Alpha,S);
    Mult(SNegHalf,S);           -- S := -(F.Alpha+G.Alpha)/2
    Translate(A,S,A_Right,R);
    Prod(B,A_Right,BA_Right);   -- BA
    Copy(A_Right,A_Left);
    Conjugate(A_Left);
    Prod(A_Left,BA_Right,C);    -- ABA
    Fix_Reversible(C);
  end Prod_GFG;

  procedure DProd_GFG(DF,DG: in Skew; O: in PData; DH: in out Skew) is
    --- DH := D(GFG)
    A_Right: TMat2 renames O.G_Right.L.all.Data;
    A_Left: TMat2 renames O.G_Left.L.all.Data;
    BA_Right: TMat2 renames O.FG_Right.L.all.Data;
    ---
    DA: TMat2 renames DG.L.all.Data;
    DB: TMat2 renames DF.L.all.Data;
    DC: TMat2 renames DH.L.all.Data;
    ---
    S: Scalar;
    T,DG_Left: Skew;
    pragma Warnings (Off,T);
    pragma Warnings (Off,DG_Left);
    Tmp: TMat2 renames T.L.all.Data;
    DA_Left: TMat2 renames DG_Left.L.all.Data;
  begin
    if not (DF.Alpha=O.F.Alpha and DG.Alpha=O.G.Alpha) then raise Domain_Error; end if;
    DH.Det1 := False;
    DH.Alpha := DF.Alpha+2*DG.Alpha;
    ---
    Prod(DB,A_Right,Tmp);
    Prod(A_Left,Tmp,DC);         -- A(DB)A
    ---
    Assign(O.FG_Right.Alpha,S);
    Mult(SHalf,S);               -- S := (F.Alpha+G.Alpha)/2
    Translate(DA,S,DA_Left,O.R);
    Prod(DA_Left,BA_Right,Tmp);  -- DA(BA)
    Add(Tmp,DC);
    Conjugate(Tmp);              -- (AB)DA
    Add(Tmp,DC);                 -- D(ABA)=A(DB)A+DA(BA)+(AB)DA
    Fix_Reversible(DC);
  end DProd_GFG;

  procedure ShowEigenA10(G: in Skew; UFac,VFac: in Scalar) is
    --- M := A0(1/2)*Lambda  in ABCD format
    GL: TMat2 renames G.L.all.Data;
    Tmp,Lambda11,Lambda22: Scalar;
    M: SMat2;
  begin
    Neg(VFac,Tmp);
    Sqrt(Tmp,Lambda11);
    Neg(UFac,Tmp);
    Sqrt(Tmp,Lambda22);
    Neg(Lambda22);
    for J in Index4 loop
      Val(GL(J),SHalf,M(J));
    end loop;
    TUVS2ABCD(M);         --- A0(1/2)
    Mult(Lambda11,M(JA));
    Mult(Lambda11,M(JC));
    Mult(Lambda22,M(JB));
    Mult(Lambda22,M(JD));
    ShowEigenVal("A10 ",M,G.Det1);
  end ShowEigenA10;

  procedure ShowCommutator(F,G: in Skew) is
    --- Commutator-1 at (1+InvGold)/2
    R1,R2: Radius;
    SA,SB: Scalar;
    A0,B0,Tmp1,Tmp2,C: SMat2;
  begin
    if not (F.Alpha=GOne and G.Alpha=Inv_Gold) then raise Not_Implemented; end if;
    if not (F.Det1 and G.Det1) then raise Not_Implemented; end if;
    Copy(SNegHalf,SA);
    InvGold(SB);
    Mult(SHalf,SB);
    Val(G,SA,A0);
    Val(F,SB,B0);
    TUVS2ABCD(A0);
    TUVS2ABCD(B0);
    Prod(B0,A0,Tmp1);
    ---
    Neg(SA);
    Neg(SB);
    Val(G,SA,A0);
    Val(F,SB,B0);
    TUVS2ABCD(A0);
    TUVS2ABCD(B0);
    Inv(A0);
    Inv(B0);
    Prod(B0,A0,Tmp2);
    ---
    Prod(Tmp1,Tmp2,C);
    Sub(SOne,C(JA));
    Sub(SOne,C(JD));
    R1 := MaxNorm(C(JA))+MaxNorm(C(JC));
    R2 := MaxNorm(C(JB))+MaxNorm(C(JB));
    Show1("|Commutator-1|",RMax(R1,R2));
  end ShowCommutator;

begin

  Use_FBalls := not (STrunc or else Std); --- for speed

end Taylors1.Skews;
